/*
 * dabpi_ctl - raspberry pi fm/fmhd/dab receiver board control interface
 * Copyright (C) 2014  Bjoern Biesenbach <bjoern@bjoern-b.de>
 * Copyright (C) 2016  Heiko Jehmlich <hje@jecons.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>				// Needed for SPI port
#include <sys/ioctl.h>			// Needed for SPI port
#include <linux/spi/spidev.h>	// Needed for SPI port
#include <wiringPi.h>			// Needed for GPIO reset

#include "si46xx.h"

#define SI46XX_RESET 	4

#define SPI_DEVICE		"/dev/spidev0.0"
#define SPI_SPEED		10000000
#define SPI_BITS		8
#define SPI_DELAY		0

int spi_fd;

void spi_init() {
	uint8_t mode = SPI_MODE_0;
	uint8_t bits = SPI_BITS;
	uint32_t speed = SPI_SPEED;

	spi_fd = open(SPI_DEVICE, O_RDWR);
	if (spi_fd < 0) {
		printf("error open: %s\n", strerror(errno));
	}

	if (ioctl(spi_fd, SPI_IOC_RD_MODE, &mode) < 0) {
		printf("SPI_IOC_RD_MODE Change failure: %s\n", strerror(errno));
	}
	if (ioctl(spi_fd, SPI_IOC_WR_MODE, &mode) < 0) {
		printf("SPI_IOC_WR_MODE Change failure: %s\n", strerror(errno));
	}

	if (ioctl(spi_fd, SPI_IOC_RD_BITS_PER_WORD, &bits) < 0) {
		printf("SPI_IOC_RD_BITS_PER_WORD Change failure: %s\n", strerror(errno));
	}
	if (ioctl(spi_fd, SPI_IOC_WR_BITS_PER_WORD, &bits) < 0) {
		printf("SPI_IOC_WR_BITS_PER_WORD Change failure: %s\n", strerror(errno));
	}

	if (ioctl(spi_fd, SPI_IOC_RD_MAX_SPEED_HZ, &speed) < 0) {
		printf("SPI_IOC_RD_MAX_SPEED_HZ Change failure: %s\n", strerror(errno));
	}
	if (ioctl(spi_fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed) < 0) {
		printf("SPI_IOC_WR_MAX_SPEED_HZ Change failure: %s\n", strerror(errno));
	}
}

int spi(uint8_t *data, int len) {
	struct spi_ioc_transfer spi;
	memset(&spi, 0, sizeof(spi));
	spi.tx_buf = (unsigned long) data;
	spi.rx_buf = (unsigned long) data;
	spi.len = len;
	spi.speed_hz = SPI_SPEED;
	spi.bits_per_word = SPI_BITS;
	spi.delay_usecs = SPI_DELAY;
	int ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &spi);
	if (ret < 0) {
		printf("SPI_IOC_MESSAGE ioctl error: %s\n", strerror(errno));
	}
	msleep(1);
	return ret;
}

void si46xx_flash_spi_init()
{
	uint8_t data[36];
	data[0] = SI46XX_FLASH_LOAD;  
	data[1] = 0x10;  //select FLASH_SET_PROP_LIST sub command
	data[2] = 0x00;
	data[3] = 0x00;    
	data[4] = 0x01;  //select SPI_CLOCK_FREQ_KHZ (0x0001)
	data[5] = 0x00;   
	data[6] = 0xb8;  //and set to 3000 (kHz)
	data[7] = 0x0b;
	data[8] = 0x03;  //SPI_MODE (0x0003)
	data[9] = 0x00;
	data[10] = 0x03; //set to mode 3
	data[11] = 0x00;
	data[12] = 0x01; //READ_CMD (0x0101)
	data[13] = 0x01;
	data[14] = 0x03; //set to 0x03
	data[15] = 0x00;
	data[16] = 0x02; //HIGH_SPEED_READ_CMD (0x0102)
	data[17] = 0x01;
	data[18] = 0x0b; //set to 0x0b
	data[19] = 0x00; 
	data[20] = 0x03; //HIGH_SPEED_MAX_READ_FREQ_MHz (0x0103)
	data[21] = 0x01;
	data[22] = 0x19; //set to 25 MHz (unavailable = 0x00)
	data[23] = 0x00; 
	data[24] = 0x01; //WRITE_CMD
	data[25] = 0x02;
	data[26] = 0x02; //set to 0x02;
	data[27] = 0x00;
	data[28] = 0x02; //ERASE_SECTOR_CMD (0x0202)
	data[29] = 0x02;
	data[30] = 0x20; //set to 0x20;
	data[31] = 0x00;
	data[32] = 0x02; //ERASE_CHIP_CMD (0x0204)
	data[33] = 0x04;
	data[34] = 0xc7; //set to 0xc7
	data[35] = 0x00;

	
	spi(data,36);
	si46xx_reply("FLASH SPI INIT");
}

void si46xx_bootflash_erase()
{
	uint8_t data[4];
	data[0] = SI46XX_FLASH_LOAD;
	data[1] = 0xff;  //select sub command chip erase
	data[2] = 0xde;
	data[3] = 0xc0;
	spi(data,4);
	msleep(500);
	si46xx_reply("CHIP ERASE");
}

void si46xx_write_file_to_bootflash(const char *filename,uint32_t address) {
	FILE *fp;
	size_t result;
	unsigned long bytes_written;
	unsigned long start_address;
	bytes_written = 0;
	start_address = address;


	fp = fopen(filename, "rb");
	if (fp == NULL) {
		printf("file error %s\r\n", filename);
		return;
	} 

    
	uint8_t data[4000 + 16];
	
	while (1) {
		//read the bytes from the file. Nr of bytes read is in result
		result = fread(data + 16, 1, 4000, fp);
		if (result == 0) 
			break;	
		data[0] = SI46XX_FLASH_LOAD;
		data[1] = 0xf0; //select block write with no errror checking 
		data[2] = 0x0c; //maybe do this better when producing something in series 
		data[3] = 0xed; //to make sure update service is not ill performed
		data[4] = 0x00; //arg0 .. arg3 (AN649, p. 383)
		data[5] = 0x00;
		data[6] = 0x00;
		data[7] = 0x00; 
		data[8] = address & 0xFF;           //address byte [7:0]
		data[9] = (address >> 8) & 0xFF;    //address byte [15:8]
		data[10] = (address >> 16) & 0xFF;  //address byte [23:16]
		data[11] = (address >> 24) & 0xFF;  //address byte [31:14]
		data[12] = result & 0xFF;           //block length byte [7:0]
		data[13] = (result >> 8) & 0xFF;    //block length byte [15:8]
		data[14] = (result >> 16) & 0xFF;   //block length byte [23:16]
		data[15] = (result >> 24) & 0xFF;   //block length byte [31:24]
		spi(data, result + 16);
		address += (uint32_t)result;
		bytes_written += result;
		msleep(5);
	    si46xx_reply("WRITE_FILE_TO_BOOTFLASH");
	}
	printf("Start address = 0x%x End address = 0x%x Bytes written = %d\r\n",start_address,address,address - start_address);
	fclose(fp);
}

int si46xx_reply(const char *log) {
	uint8_t data[5];
	uint8_t timeout = 200; //make this longer for the flash / flash load commands
                            //in total 2 seconds then
	while (timeout--) {
		data[0] = 0;
		spi(data, 5);
		if (data[1] & 0x80) { // CTS ?
			hexDump(log, data, 5);
			return 1;
		}
		msleep(10);
	}
	printf("timeout on %s\r\n", log);
	return 0;
}

void si46xx_boot() {
	uint8_t data[2];
	data[0] = SI46XX_BOOT;
	data[1] = 0;
	spi(data, 2);
	msleep(500); // 63ms at analog fm, 198ms at DAB
	si46xx_reply("BOOT");
}

void si46xx_hostload(const char *filename) {
	FILE *fp;
	size_t result;

	uint8_t init[2];
	init[0] = SI46XX_LOAD_INIT;
	init[1] = 0;
	spi(init, 2);
	si46xx_reply("LOAD_INIT");

	fp = fopen(filename, "rb");
	if (fp == NULL) {
		printf("file error %s\r\n", filename);
		return;
	}

	int loop = 0;
	uint8_t data[4092 + 4];
	while (1) {
		data[0] = SI46XX_HOST_LOAD;
		data[1] = 0;
		data[2] = 0;
		data[3] = 0;
		result = fread(data + 4, 1, 4092, fp);
		if (result == 0) {
			break;
		}
		spi(data, result + 4);
		msleep(5);
		// printf("loop %d, count %d\r\n", loop++, result);
		si46xx_reply("HOST_LOAD");
	}
	fclose(fp);
}


void si46xx_flashload_image(uint32_t address)
{
	uint8_t init[2];
	init[0] = SI46XX_LOAD_INIT;
	init[1] = 0;
	spi(init, 2);
	si46xx_reply("LOAD_INIT");
	msleep(50);
	printf("FLASH LOAD ADDRESS 0x%x\r\n",address);
	uint8_t data[12];
	data[0] = SI46XX_FLASH_LOAD;  //caution - there is inconsistency in the AN 649
	data[1] = 0x00;  //flash load needs the 12 byte format, not the 8 byte format !
	data[2] = 0x00;
	data[3] = 0x00;
	data[4] = address & 0xff;
	data[5] = (address >> 8) & 0xff;
	data[6] = (address >> 16) & 0xff;
	data[7] = (address >> 24) & 0xff;
	data[8] = 0x00;
	data[9] = 0x00;
	data[10] = 0x00;
	data[11] = 0x00;
	spi(data,12);
	si46xx_reply("FLASH_LOAD");
}

void si46xx_powerup() {
	uint8_t data[16];

	data[0] = SI46XX_POWER_UP; // COMMAND
	data[1] = 0x00; // ARG1 disable CTSIEN
	data[2] = 0x17; //(1 << 4) | (7 << 0); // ARG2 CLK_MODE=0x1 TR_SIZE=0x7
	data[3] = 0x48; // ARG3 IBIAS=0x48
	data[4] = 0x00; // ARG4 XTAL
	data[5] = 0xF9; // ARG5 XTAL // F8
	data[6] = 0x24; // ARG6 XTAL
	data[7] = 0x01; // ARG7 XTAL 19.2MHz
	data[8] = 0x1F; // ARG8 CTUN
	data[9] = 0x00 | (1 << 4); // ARG9
	data[10] = 0x00; // ARG10
	data[11] = 0x00; // ARG11
	data[12] = 0x00; // ARG12
	data[13] = 0x00; // ARG13 IBIAS_RUN
	data[14] = 0x00; // ARG14
	data[15] = 0x00; // ARG15

	spi(data, 16);
	si46xx_reply("POWER_UP");
}

void si46xx_get_partinfo() {
	uint8_t data[23];
	data[0] = SI46XX_GET_PART_INFO;
	data[1] = 0;
	spi(data, 2);
	msleep(5);

	data[0] = 0;
	spi(data, 23);
	hexDump("GET_PART_INFO", data, 23);
}

void si46xx_get_sys_state() {
	uint8_t data[7];
	data[0] = SI46XX_GET_SYS_STATE;
	data[1] = 0;
	spi(data, 2);
	msleep(5);

	data[0] = 0;
	spi(data, 7);
	hexDump("GET_SYS_STATE", data, 7);

	printf("Current mode: \n");
	switch (data[5]) {
	case 0:
		printf("Bootloader is active\n");
		break;
	case 1:
		printf("FMHD is active\n");
		break;
	case 2:
		printf("DAB is active\n");
		break;
	case 3:
		printf("TDMB or data only DAB image is active\n");
		break;
	case 4:
		printf("FMHD is active\n");
		break;
	case 5:
		printf("AMHD is active\n");
		break;
	case 6:
		printf("AMHD Demod is active\n");
		break;
	default:
		break;
	}
}

void si46xx_set_property(uint16_t name, uint16_t value) {
	uint8_t data[6];
	data[0] = SI46XX_SET_PROPERTY;
	data[1] = 0;
	data[2] = name & 0xFF;
	data[3] = (name >> 8) & 0xFF;
	data[4] = value & 0xFF;
	data[5] = (value >> 8) & 0xFF;
	spi(data, 6);
	si46xx_reply("SET_PROPERTY");
}

void si46xx_reset() {
	wiringPiSetup();
	pinMode(SI46XX_RESET, OUTPUT);

	/* reset si46xx  */
	digitalWrite(SI46XX_RESET, 0);
	msleep(10);
	digitalWrite(SI46XX_RESET, 1);
	msleep(10);
}
